"""A runner for the Python code generated from Safe-DS programs."""

from .server._pipeline_manager import file_mtime, memoized_call, save_placeholder

__all__ = [
    "file_mtime",
    "memoized_call",
    "save_placeholder",
]



"""Module that contains functions for creating and validating messages exchanged with the vscode extension."""

from __future__ import annotations

import dataclasses
import json
from dataclasses import dataclass
from typing import Any

message_type_program = "program"
message_type_placeholder_query = "placeholder_query"
message_type_placeholder_type = "placeholder_type"
message_type_placeholder_value = "placeholder_value"
message_type_runtime_error = "runtime_error"
message_type_runtime_progress = "runtime_progress"
message_type_shutdown = "shutdown"

message_types = [
    message_type_program,
    message_type_placeholder_query,
    message_type_placeholder_type,
    message_type_placeholder_value,
    message_type_runtime_error,
    message_type_runtime_progress,
    message_type_shutdown,
]


@dataclass(frozen=True)
class Message:
    """
    A message object, which is exchanged between the runner and the VS Code extension.

    Parameters
    ----------
    type : str
        Type that identifies the kind of message.
    id : str
        ID that identifies the execution where this message belongs to.
    data : Any
        Message data section. Differs between message types.
    """

    type: str
    id: str
    data: Any

    @staticmethod
    def from_dict(d: dict[str, Any]) -> Message:
        """
        Create a new Message object from a dictionary.

        Parameters
        ----------
        d : dict[str, Any]
            Dictionary which should contain all needed fields.

        Returns
        -------
        Message
            Dataclass which contains information copied from the provided dictionary.
        """
        return Message(**d)

    def to_dict(self) -> dict[str, Any]:
        """
        Convert this dataclass to a dictionary.

        Returns
        -------
        dict[str, Any]
            Dictionary containing all the fields which are part of this dataclass.
        """
        return dataclasses.asdict(self)


@dataclass(frozen=True)
class MessageDataProgram:
    """
    Message data for a program message.

    Parameters
    ----------
    code : dict[str, dict[str, str]]
        A dictionary containing the code needed for executed,
        in a virtual filesystem. Keys of the outer dictionary are the module path, keys of the inner dictionary are the
        module name. The values of the inner dictionary is the python code for each module.
    main : ProgramMainInformation
        Information where the main pipeline (the pipeline to be executed) is located.
    """

    code: dict[str, dict[str, str]]
    main: ProgramMainInformation

    @staticmethod
    def from_dict(d: dict[str, Any]) -> MessageDataProgram:
        """
        Create a new MessageDataProgram object from a dictionary.

        Parameters
        ----------
        d : dict[str, Any]
            Dictionary which should contain all needed fields.

        Returns
        -------
        MessageDataProgram
            Dataclass which contains information copied from the provided dictionary.
        """
        return MessageDataProgram(d["code"], ProgramMainInformation.from_dict(d["main"]))

    def to_dict(self) -> dict[str, Any]:
        """
        Convert this dataclass to a dictionary.

        Returns
        -------
        dict[str, Any]
            Dictionary containing all the fields which are part of this dataclass.
        """
        return dataclasses.asdict(self)  # pragma: no cover


@dataclass(frozen=True)
class ProgramMainInformation:
    """
    Information that can be used to locate a pipeline.

    Parameters
    ----------
    modulepath : str
        Path, where the main module is located.
    module : str
        Safe-DS module name.
    pipeline : str
        Safe-DS pipeline name.
    """

    modulepath: str
    module: str
    pipeline: str

    @staticmethod
    def from_dict(d: dict[str, Any]) -> ProgramMainInformation:
        """
        Create a new ProgramMainInformation object from a dictionary.

        Parameters
        ----------
        d : dict[str, Any]
            Dictionary which should contain all needed fields.

        Returns
        -------
        ProgramMainInformation
            Dataclass which contains information copied from the provided dictionary.
        """
        return ProgramMainInformation(**d)

    def to_dict(self) -> dict[str, Any]:
        """
        Convert this dataclass to a dictionary.

        Returns
        -------
        dict[str, Any]
            Dictionary containing all the fields which are part of this dataclass.
        """
        return dataclasses.asdict(self)  # pragma: no cover


@dataclass(frozen=True)
class QueryWindow:
    """
    Information that is used to create a subset of the data of a placeholder.

    Parameters
    ----------
    begin : int | None
        Index of the first entry that should be sent. May be present if a windowed query is required.
    size : int | None
        Max. amount of entries that should be sent. May be present if a windowed query is required.
    """

    begin: int | None = None
    size: int | None = None

    @staticmethod
    def from_dict(d: dict[str, Any]) -> QueryWindow:
        """
        Create a new QueryWindow object from a dictionary.

        Parameters
        ----------
        d : dict[str, Any]
            Dictionary which should contain all needed fields.

        Returns
        -------
        QueryWindow
            Dataclass which contains information copied from the provided dictionary.
        """
        return QueryWindow(**d)

    def to_dict(self) -> dict[str, Any]:
        """
        Convert this dataclass to a dictionary.

        Returns
        -------
        dict[str, Any]
            Dictionary containing all the fields which are part of this dataclass.
        """
        return dataclasses.asdict(self)  # pragma: no cover


@dataclass(frozen=True)
class MessageQueryInformation:
    """
    Information used to query a placeholder with optional window bounds. Only complex types like tables are affected by window bounds.

    Parameters
    ----------
    name : str
        Placeholder name that is queried.
    window : QueryWindow
        Window bounds for requesting only a subset of the available data.
    """

    name: str
    window: QueryWindow = dataclasses.field(default_factory=QueryWindow)

    @staticmethod
    def from_dict(d: dict[str, Any]) -> MessageQueryInformation:
        """
        Create a new MessageQueryInformation object from a dictionary.

        Parameters
        ----------
        d : dict[str, Any]
            Dictionary which should contain all needed fields.

        Returns
        -------
        MessageQueryInformation
            Dataclass which contains information copied from the provided dictionary.
        """
        return MessageQueryInformation(name=d["name"], window=QueryWindow.from_dict(d["window"]))

    def to_dict(self) -> dict[str, Any]:
        """
        Convert this dataclass to a dictionary.

        Returns
        -------
        dict[str, Any]
            Dictionary containing all the fields which are part of this dataclass.
        """
        return dataclasses.asdict(self)  # pragma: no cover


def create_placeholder_description(name: str, type_: str) -> dict[str, str]:
    """
    Create the message data of a placeholder description message containing only name and type.

    Parameters
    ----------
    name : str
        Name of the placeholder.
    type_ : str
        Type of the placeholder.

    Returns
    -------
    dict[str, str]
        Message data of "placeholder_type" messages.
    """
    return {"name": name, "type": type_}


def create_placeholder_value(placeholder_query: MessageQueryInformation, type_: str, value: Any) -> dict[str, Any]:
    """
    Create the message data of a placeholder value message containing name, type and the actual value.

    If the query only requests a subset of the data and the placeholder type supports this,
    the response will contain only a subset and the information about the subset.

    Parameters
    ----------
    placeholder_query : MessageQueryInformation
        Query of the placeholder.
    type_ : str
        Type of the placeholder.
    value : Any
        Value of the placeholder.

    Returns
    -------
    dict[str, str]
        Message data of "placeholder_value" messages.
    """
    import safeds.data.tabular.containers

    message: dict[str, Any] = {"name": placeholder_query.name, "type": type_}
    # Start Index >= 0
    start_index = max(placeholder_query.window.begin if placeholder_query.window.begin is not None else 0, 0)
    # End Index >= Start Index
    end_index = (
        (start_index + max(placeholder_query.window.size, 0)) if placeholder_query.window.size is not None else None
    )
    if isinstance(value, safeds.data.tabular.containers.Table) and (
        placeholder_query.window.begin is not None or placeholder_query.window.size is not None
    ):
        max_index = value.number_of_rows
        # End Index <= Number Of Rows
        end_index = min(end_index, value.number_of_rows) if end_index is not None else None
        value = value.slice_rows(start=start_index, end=end_index)
        window_information: dict[str, int] = {"begin": start_index, "size": value.number_of_rows, "max": max_index}
        message["window"] = window_information
    message["value"] = value
    return message


def create_runtime_error_description(message: str, backtrace: list[dict[str, Any]]) -> dict[str, Any]:
    """
    Create the message data of a runtime error message containing error information and a backtrace.

    Parameters
    ----------
    message : str
        Error information message.
    backtrace : list[dict[str, Any]]
        Python backtrace of the error. Each list entry represents a stack frame.

    Returns
    -------
    dict[str, Any]
        Message data of "runtime_error" messages.
    """
    return {"message": message, "backtrace": backtrace}


def create_runtime_progress_done() -> str:
    """
    Create the message data of a runtime progress message containing 'done'.

    Returns
    -------
    str
        Message data of "runtime_progress" messages.
    """
    return "done"


def parse_validate_message(message: str) -> tuple[Message | None, str | None, str | None]:
    """
    Validate the basic structure of a received message string and return a parsed message object.

    Parameters
    ----------
    message : str
        Message string, that should be in JSON format.

    Returns
    -------
    tuple[Message | None, str | None, str | None]
        A tuple containing either a message or a detailed error description and a short error message.
    """
    try:
        message_dict: dict[str, Any] = json.loads(message)
    except json.JSONDecodeError:
        return None, f"Invalid message received: {message}", "Invalid Message: not JSON"
    if "type" not in message_dict:
        return None, f"No message type specified in: {message}", "Invalid Message: no type"
    elif "id" not in message_dict:
        return None, f"No message id specified in: {message}", "Invalid Message: no id"
    elif "data" not in message_dict:
        return None, f"No message data specified in: {message}", "Invalid Message: no data"
    elif not isinstance(message_dict["type"], str):
        return None, f"Message type is not a string: {message}", "Invalid Message: invalid type"
    elif not isinstance(message_dict["id"], str):
        return None, f"Message id is not a string: {message}", "Invalid Message: invalid id"
    else:
        return Message.from_dict(message_dict), None, None


def validate_program_message_data(message_data: dict[str, Any] | str) -> tuple[MessageDataProgram | None, str | None]:
    """
    Validate the message data of a program message.

    Parameters
    ----------
    message_data : dict[str, Any] | str
        Message data dictionary or string.

    Returns
    -------
    tuple[MessageDataProgram | None, str | None]
        A tuple containing either a validated message data object or an error message.
    """
    if not isinstance(message_data, dict):
        return None, "Message data is not a JSON object"
    elif "code" not in message_data:
        return None, "No 'code' parameter given"
    elif "main" not in message_data:
        return None, "No 'main' parameter given"
    elif (
        not isinstance(message_data["main"], dict)
        or "modulepath" not in message_data["main"]
        or "module" not in message_data["main"]
        or "pipeline" not in message_data["main"]
        or len(message_data["main"]) != 3
    ):
        return None, "Invalid 'main' parameter given"
    elif not isinstance(message_data["code"], dict):
        return None, "Invalid 'code' parameter given"
    else:
        code: dict = message_data["code"]
        for key in code:
            if not isinstance(code[key], dict):
                return None, "Invalid 'code' parameter given"
            next_dict: dict = code[key]
            for next_key in next_dict:
                if not isinstance(next_dict[next_key], str):
                    return None, "Invalid 'code' parameter given"
        return MessageDataProgram.from_dict(message_data), None


def validate_placeholder_query_message_data(
    message_data: dict[str, Any] | str,
) -> tuple[MessageQueryInformation | None, str | None]:
    """
    Validate the message data of a placeholder query message.

    Parameters
    ----------
    message_data : dict[str, Any] | str
        Message data dictionary or string.

    Returns
    -------
    tuple[MessageQueryInformation | None, str | None]
        A tuple containing either the validated message data or an error message.
    """
    if not isinstance(message_data, dict):
        return None, "Message data is not a JSON object"
    elif "name" not in message_data:
        return None, "No 'name' parameter given"
    elif (
        "window" in message_data
        and "begin" in message_data["window"]
        and not isinstance(message_data["window"]["begin"], int)
    ):
        return None, "Invalid 'window'.'begin' parameter given"
    elif (
        "window" in message_data
        and "size" in message_data["window"]
        and not isinstance(message_data["window"]["size"], int)
    ):
        return None, "Invalid 'window'.'size' parameter given"
    else:
        return MessageQueryInformation.from_dict(message_data), None



"""Module containing the server, endpoints and utility functions."""

import asyncio
import json
import logging
import sys

import hypercorn.asyncio
import quart.app

from ._json_encoder import SafeDsEncoder
from ._messages import (
    Message,
    create_placeholder_value,
    message_type_placeholder_value,
    message_types,
    parse_validate_message,
    validate_placeholder_query_message_data,
    validate_program_message_data,
)
from ._pipeline_manager import PipelineManager


def create_flask_app() -> quart.app.Quart:
    """
    Create a quart app, that handles all requests.

    Returns
    -------
    quart.app.Quart
        App.
    """
    return quart.app.Quart(__name__)


class SafeDsServer:
    """Server containing the flask app, websocket handler and endpoints."""

    def __init__(self) -> None:
        """Create a new server object."""
        self.app_pipeline_manager = PipelineManager()
        self.app = create_flask_app()
        self.app.config["pipeline_manager"] = self.app_pipeline_manager
        self.app.websocket("/WSMain")(SafeDsServer.ws_main)

    def listen(self, port: int) -> None:
        """
        Listen on the specified port for incoming connections to the runner.

        Parameters
        ----------
        port : int
            Port to listen on
        """
        logging.info("Starting Safe-DS Runner on port %s", str(port))
        serve_config = hypercorn.config.Config()
        # Only bind to host=127.0.0.1. Connections from other devices should not be accepted
        serve_config.bind = f"127.0.0.1:{port}"
        serve_config.websocket_ping_interval = 25.0
        event_loop = asyncio.get_event_loop()
        event_loop.run_until_complete(hypercorn.asyncio.serve(self.app, serve_config))
        event_loop.run_forever()  # pragma: no cover

    @staticmethod
    async def ws_main() -> None:
        """Handle websocket requests to the WSMain endpoint and delegates with the required objects."""
        await SafeDsServer._ws_main(quart.websocket, quart.current_app.config["pipeline_manager"])

    @staticmethod
    async def _ws_main(ws: quart.Websocket, pipeline_manager: PipelineManager) -> None:
        """
        Handle websocket requests to the WSMain endpoint.

        This function handles the bidirectional communication between the runner and the VS Code extension.

        Parameters
        ----------
        ws : quart.Websocket
            Connection
        pipeline_manager : PipelineManager
            Pipeline Manager
        """
        logging.debug("Request to WSRunProgram")
        output_queue: asyncio.Queue = asyncio.Queue()
        pipeline_manager.connect(output_queue)
        foreground_handler = asyncio.create_task(SafeDsServer._ws_main_foreground(ws, pipeline_manager, output_queue))
        background_handler = asyncio.create_task(SafeDsServer._ws_main_background(ws, output_queue))
        await asyncio.gather(foreground_handler, background_handler)

    @staticmethod
    async def _ws_main_foreground(
        ws: quart.Websocket,
        pipeline_manager: PipelineManager,
        output_queue: asyncio.Queue,
    ) -> None:
        while True:
            # This would be a JSON message
            received_message: str = await ws.receive()
            logging.debug("Received Message: %s", received_message)
            received_object, error_detail, error_short = parse_validate_message(received_message)
            if received_object is None:
                logging.error(error_detail)
                await output_queue.put(None)
                pipeline_manager.disconnect(output_queue)
                await ws.close(code=1000, reason=error_short)
                return
            match received_object.type:
                case "shutdown":
                    logging.debug("Requested shutdown...")
                    pipeline_manager.shutdown()
                    sys.exit(0)
                case "program":
                    program_data, invalid_message = validate_program_message_data(received_object.data)
                    if program_data is None:
                        logging.error("Invalid message data specified in: %s (%s)", received_message, invalid_message)
                        await output_queue.put(None)
                        pipeline_manager.disconnect(output_queue)
                        await ws.close(code=1000, reason=invalid_message)
                        return
                    # This should only be called from the extension as it is a security risk
                    pipeline_manager.execute_pipeline(program_data, received_object.id)
                case "placeholder_query":
                    # For this query, a response can be directly sent to the requesting connection
                    placeholder_query_data, invalid_message = validate_placeholder_query_message_data(
                        received_object.data,
                    )
                    if placeholder_query_data is None:
                        logging.error("Invalid message data specified in: %s (%s)", received_message, invalid_message)
                        await output_queue.put(None)
                        pipeline_manager.disconnect(output_queue)
                        await ws.close(code=1000, reason=invalid_message)
                        return
                    placeholder_type, placeholder_value = pipeline_manager.get_placeholder(
                        received_object.id,
                        placeholder_query_data.name,
                    )
                    # send back a value message
                    if placeholder_type is not None:
                        try:
                            await send_message(
                                ws,
                                Message(
                                    message_type_placeholder_value,
                                    received_object.id,
                                    create_placeholder_value(
                                        placeholder_query_data,
                                        placeholder_type,
                                        placeholder_value,
                                    ),
                                ),
                            )
                        except TypeError as _encoding_error:
                            # if the value can't be encoded send back that the value exists but is not displayable
                            await send_message(
                                ws,
                                Message(
                                    message_type_placeholder_value,
                                    received_object.id,
                                    create_placeholder_value(
                                        placeholder_query_data,
                                        placeholder_type,
                                        "<Not displayable>",
                                    ),
                                ),
                            )
                    else:
                        # Send back empty type / value, to communicate that no placeholder exists (yet)
                        # Use name from query to allow linking a response to a request on the peer
                        await send_message(
                            ws,
                            Message(
                                message_type_placeholder_value,
                                received_object.id,
                                create_placeholder_value(placeholder_query_data, "", ""),
                            ),
                        )
                case _:
                    if received_object.type not in message_types:
                        logging.warning("Invalid message type: %s", received_object.type)

    @staticmethod
    async def _ws_main_background(ws: quart.Websocket, output_queue: asyncio.Queue) -> None:
        while True:
            encoded_message = await output_queue.get()
            if encoded_message is None:
                return
            await ws.send(encoded_message)


async def send_message(connection: quart.Websocket, message: Message) -> None:
    """
    Send a message to the provided websocket connection (to the VS Code extension).

    Parameters
    ----------
    connection : quart.Websocket
        Connection that should receive the message.
    message : Message
        Object that will be sent.
    """
    message_encoded = json.dumps(message.to_dict(), cls=SafeDsEncoder)
    await connection.send(message_encoded)



"""Module that contains the memoization logic and stats."""

import dataclasses
import logging
import sys
import time
from collections.abc import Callable
from dataclasses import dataclass
from typing import Any, TypeAlias

MemoizationKey: TypeAlias = tuple[str, tuple[Any], tuple[Any]]


@dataclass(frozen=True)
class MemoizationStats:
    """
    Statistics calculated for every memoization call.

    Parameters
    ----------
    access_timestamps
        Absolute timestamp since the unix epoch of the last access to the memoized value in nanoseconds
    lookup_times
        Duration the lookup of the value took in nanoseconds (key comparison + IPC)
    computation_times
        Duration the computation of the value took in nanoseconds
    memory_sizes
        Amount of memory the memoized value takes up in bytes
    """

    access_timestamps: list[int] = dataclasses.field(default_factory=list)
    lookup_times: list[int] = dataclasses.field(default_factory=list)
    computation_times: list[int] = dataclasses.field(default_factory=list)
    memory_sizes: list[int] = dataclasses.field(default_factory=list)

    def update_on_hit(self, access_timestamp: int, lookup_time: int) -> None:
        """
        Update the memoization stats on a cache hit.

        Parameters
        ----------
        access_timestamp
            Timestamp when this value was last accessed
        lookup_time
            Duration the comparison took in nanoseconds
        """
        self.access_timestamps.append(access_timestamp)
        self.lookup_times.append(lookup_time)

    def update_on_miss(self, access_timestamp: int, lookup_time: int, computation_time: int, memory_size: int) -> None:
        """
        Update the memoization stats on a cache miss.

        Parameters
        ----------
        access_timestamp
            Timestamp when this value was last accessed
        lookup_time
            Duration the comparison took in nanoseconds
        computation_time
            Duration the computation of the new value took in nanoseconds
        memory_size
            Memory the newly computed value takes up in bytes
        """
        self.access_timestamps.append(access_timestamp)
        self.lookup_times.append(lookup_time)
        self.computation_times.append(computation_time)
        self.memory_sizes.append(memory_size)

    def __str__(self) -> str:
        """
        Summarizes stats contained in this object.

        Returns
        -------
        Summary of stats
        """
        return (  # pragma: no cover
            f"Last access: {self.access_timestamps}, computation time: {self.computation_times}, lookup time:"
            f" {self.lookup_times}, memory size: {self.memory_sizes}"
        )


class MemoizationMap:
    """
    The memoization map handles memoized function calls.

    This contains looking up stored values, computing new values if needed and calculating and updating statistics.
    """

    def __init__(
        self,
        map_values: dict[MemoizationKey, Any],
        map_stats: dict[str, MemoizationStats],
    ):
        """
        Create a new memoization map using a value store dictionary and a stats dictionary.

        Parameters
        ----------
        map_values
            Value store dictionary
        map_stats
            Stats dictionary
        """
        self._map_values: dict[MemoizationKey, Any] = map_values
        self._map_stats: dict[str, MemoizationStats] = map_stats

    def memoized_function_call(
        self,
        function_name: str,
        function_callable: Callable,
        parameters: list[Any],
        hidden_parameters: list[Any],
    ) -> Any:
        """
        Handle a memoized function call.

        Looks up the stored value, determined by function name, parameters and hidden parameters and returns it if found.
        If no value is found, computes the value using the provided callable and stores it in the map.
        Every call to this function will update the memoization stats.

        Parameters
        ----------
        function_name
            Fully qualified function name
        function_callable
            Function that is called and memoized if the result was not found in the memoization map
        parameters
            List of parameters passed to the function
        hidden_parameters
            List of hidden parameters for the function. This is used for memoizing some impure functions.

        Returns
        -------
        The result of the specified function, if any exists
        """
        access_timestamp = time.time_ns()

        # Lookup memoized value
        lookup_time_start = time.perf_counter_ns()
        key = self._create_memoization_key(function_name, parameters, hidden_parameters)
        memoized_value = self._lookup_value(key)
        lookup_time = time.perf_counter_ns() - lookup_time_start

        # Hit
        if memoized_value is not None:
            self._update_stats_on_hit(function_name, access_timestamp, lookup_time)
            return memoized_value

        # Miss
        computation_time_start = time.perf_counter_ns()
        computed_value = self._compute_and_memoize_value(key, function_callable, parameters)
        computation_time = time.perf_counter_ns() - computation_time_start
        memory_size = _get_size_of_value(computed_value)

        self._update_stats_on_miss(
            function_name,
            access_timestamp,
            lookup_time,
            computation_time,
            memory_size,
        )

        logging.info(
            "New memoization stats for %s: (access_timestamp=%s, lookup_time=%s, computation_time=%s, memory_size=%s)",
            key[0],
            access_timestamp,
            lookup_time,
            computation_time,
            memory_size,
        )

        return computed_value

    def _create_memoization_key(
        self,
        function_name: str,
        parameters: list[Any],
        hidden_parameters: list[Any],
    ) -> MemoizationKey:
        """
        Convert values provided to a memoized function call to a memoization key.

        Parameters
        ----------
        function_name
            Fully qualified function name
        parameters
            List of parameters passed to the function
        hidden_parameters
            List of parameters not passed to the function

        Returns
        -------
        A memoization key, which contains the lists converted to tuples
        """
        return function_name, _convert_list_to_tuple(parameters), _convert_list_to_tuple(hidden_parameters)

    def _lookup_value(self, key: MemoizationKey) -> Any | None:
        """
        Lookup a potentially existing value from the memoization cache.

        Parameters
        ----------
        key
            Memoization Key

        Returns
        -------
        The value corresponding to the provided memoization key, if any exists.
        """
        return self._map_values.get(key)

    def _compute_and_memoize_value(
        self,
        key: MemoizationKey,
        function_callable: Callable,
        parameters: list[Any],
    ) -> Any:
        """
        Memoize a new function call and return computed the result.

        Parameters
        ----------
        key
            Memoization Key
        function_callable
            Function that will be called
        parameters
            List of parameters passed to the function

        Returns
        -------
        The newly computed value corresponding to the provided memoization key
        """
        result = function_callable(*parameters)
        self._map_values[key] = result
        return result

    def _update_stats_on_hit(self, function_name: str, access_timestamp: int, lookup_time: int) -> None:
        """
        Update the memoization stats on a cache hit.

        Parameters
        ----------
        function_name
            Fully qualified function name
        access_timestamp
            Timestamp when this value was last accessed
        lookup_time
            Duration the comparison took in nanoseconds
        """
        stats = self._map_stats[function_name]
        stats.update_on_hit(access_timestamp, lookup_time)

        # This assignment is required for multiprocessing, see
        # https://docs.python.org/3.11/library/multiprocessing.html#proxy-objects
        self._map_stats[function_name] = stats

    def _update_stats_on_miss(
        self,
        function_name: str,
        access_timestamp: int,
        lookup_time: int,
        computation_time: int,
        memory_size: int,
    ) -> None:
        """
        Update the memoization stats on a cache miss.

        Parameters
        ----------
        function_name
            Fully qualified function name
        access_timestamp
            Timestamp when this value was last accessed
        lookup_time
            Duration the comparison took in nanoseconds
        computation_time
            Duration the computation of the new value took in nanoseconds
        memory_size
            Memory the newly computed value takes up in bytes
        """
        stats = self._map_stats.get(function_name)
        if stats is None:
            stats = MemoizationStats()

        stats.update_on_miss(access_timestamp, lookup_time, computation_time, memory_size)
        self._map_stats[function_name] = stats


def _convert_list_to_tuple(values: list) -> tuple:
    """
    Recursively convert a mutable list of values to an immutable tuple containing the same values, to make the values hashable.

    Parameters
    ----------
    values : list
        Values that should be converted to a tuple

    Returns
    -------
    tuple
        Converted list containing all the elements of the provided list
    """
    return tuple(_convert_list_to_tuple(value) if isinstance(value, list) else value for value in values)


def _get_size_of_value(value: Any) -> int:
    """
    Recursively calculate the memory usage of a given value.

    Parameters
    ----------
    value
        Any value of which the memory usage should be calculated.

    Returns
    -------
    Size of the provided value in bytes
    """
    size_immediate = sys.getsizeof(value)
    if isinstance(value, dict):
        return (
            sum(map(_get_size_of_value, value.keys())) + sum(map(_get_size_of_value, value.values())) + size_immediate
        )
    elif isinstance(value, frozenset | list | set | tuple):
        return sum(map(_get_size_of_value, value)) + size_immediate
    else:
        return size_immediate



"""Module that contains the infrastructure for pipeline execution in child processes."""

import asyncio
import json
import logging
import multiprocessing
import queue
import runpy
import threading
import typing
from functools import cached_property
from multiprocessing.managers import SyncManager
from pathlib import Path
from typing import Any

import stack_data

from ._memoization_map import MemoizationMap
from ._messages import (
    Message,
    MessageDataProgram,
    create_placeholder_description,
    create_runtime_error_description,
    create_runtime_progress_done,
    message_type_placeholder_type,
    message_type_runtime_error,
    message_type_runtime_progress,
)
from ._module_manager import InMemoryFinder


class PipelineManager:
    """
    A PipelineManager handles the execution of pipelines and processing of results.

    This includes launching subprocesses and the communication between the
    subprocess and the main process using a shared message queue.
    """

    def __init__(self) -> None:
        """Create a new PipelineManager object, which is lazily started, when needed."""
        self._placeholder_map: dict = {}
        self._websocket_target: list[asyncio.Queue] = []

    @cached_property
    def _multiprocessing_manager(self) -> SyncManager:
        if multiprocessing.get_start_method() != "spawn":
            multiprocessing.set_start_method("spawn", force=True)
        return multiprocessing.Manager()

    @cached_property
    def _messages_queue(self) -> queue.Queue[Message]:
        return self._multiprocessing_manager.Queue()

    @cached_property
    def _messages_queue_thread(self) -> threading.Thread:
        return threading.Thread(target=self._handle_queue_messages, daemon=True, args=(asyncio.get_event_loop(),))

    @cached_property
    def _memoization_map(self) -> MemoizationMap:
        return MemoizationMap(self._multiprocessing_manager.dict(), self._multiprocessing_manager.dict())  # type: ignore[arg-type]

    def startup(self) -> None:
        """
        Prepare the runner for running Safe-DS pipelines.

        Firstly, structures shared between processes are lazily created.
        After that a message queue handling thread is started in the main process.
        This allows receiving messages directly from one of the pipeline processes and relaying information
        directly to the websocket connection (to the VS Code extension).

        This method should not be called during the bootstrap phase of the python interpreter, as it leads to a crash.
        """
        _mq = self._messages_queue  # Initialize it here before starting a thread to avoid potential race condition
        if not self._messages_queue_thread.is_alive():
            self._messages_queue_thread.start()

    def _handle_queue_messages(self, event_loop: asyncio.AbstractEventLoop) -> None:
        """
        Relay messages from pipeline processes to the currently connected websocket endpoint.

        Should be used in a dedicated thread.

        Parameters
        ----------
        event_loop : asyncio.AbstractEventLoop
            Event Loop that handles websocket connections.
        """
        try:
            while self._messages_queue is not None:
                message = self._messages_queue.get()
                message_encoded = json.dumps(message.to_dict())
                # only send messages to the same connection once
                for connection in set(self._websocket_target):
                    asyncio.run_coroutine_threadsafe(connection.put(message_encoded), event_loop)
        except BaseException as error:  # noqa: BLE001  # pragma: no cover
            logging.warning("Message queue terminated: %s", error.__repr__())  # pragma: no cover

    def connect(self, websocket_connection_queue: asyncio.Queue) -> None:
        """
        Add a websocket connection queue to relay event messages to, which are occurring during pipeline execution.

        Parameters
        ----------
        websocket_connection_queue : asyncio.Queue
            Message Queue for a websocket connection.
        """
        self._websocket_target.append(websocket_connection_queue)

    def disconnect(self, websocket_connection_queue: asyncio.Queue) -> None:
        """
        Remove a websocket target connection queue to no longer receive messages.

        Parameters
        ----------
        websocket_connection_queue : asyncio.Queue
            Message Queue for a websocket connection to be removed.
        """
        self._websocket_target.remove(websocket_connection_queue)

    def execute_pipeline(
        self,
        pipeline: MessageDataProgram,
        execution_id: str,
    ) -> None:
        """
        Run a Safe-DS pipeline.

        Parameters
        ----------
        pipeline : MessageDataProgram
            Message object that contains the information to run a pipeline.
        execution_id : str
            Unique ID to identify this execution.
        """
        self.startup()
        if execution_id not in self._placeholder_map:
            self._placeholder_map[execution_id] = self._multiprocessing_manager.dict()
        process = PipelineProcess(
            pipeline,
            execution_id,
            self._messages_queue,
            self._placeholder_map[execution_id],
            self._memoization_map,
        )
        process.execute()

    def get_placeholder(self, execution_id: str, placeholder_name: str) -> tuple[str | None, Any]:
        """
        Get a placeholder type and value for an execution id and placeholder name.

        Parameters
        ----------
        execution_id : str
            Unique ID identifying the execution in which the placeholder was calculated.
        placeholder_name : str
            Name of the placeholder.

        Returns
        -------
        tuple[str | None, Any]
            Tuple containing placeholder type and placeholder value, or (None, None) if the placeholder does not exist.
        """
        if execution_id not in self._placeholder_map:
            return None, None
        if placeholder_name not in self._placeholder_map[execution_id]:
            return None, None
        value = self._placeholder_map[execution_id][placeholder_name]
        return _get_placeholder_type(value), value

    def shutdown(self) -> None:
        """
        Shut down the multiprocessing manager to end the used subprocess.

        This should only be called if this PipelineManager is not intended to be reused again.
        """
        self._multiprocessing_manager.shutdown()


class PipelineProcess:
    """A process that executes a Safe-DS pipeline."""

    def __init__(
        self,
        pipeline: MessageDataProgram,
        execution_id: str,
        messages_queue: queue.Queue[Message],
        placeholder_map: dict[str, Any],
        memoization_map: MemoizationMap,
    ):
        """
        Create a new process which will execute the given pipeline, when started.

        Parameters
        ----------
        pipeline : MessageDataProgram
            Message object that contains the information to run a pipeline.
        execution_id : str
            Unique ID to identify this process.
        messages_queue : queue.Queue[Message]
            A queue to write outgoing messages to.
        placeholder_map : dict[str, Any]
            A map to save calculated placeholders in.
        memoization_map : MemoizationMap
            A map to save memoizable functions in.
        """
        self._pipeline = pipeline
        self._id = execution_id
        self._messages_queue = messages_queue
        self._placeholder_map = placeholder_map
        self._memoization_map = memoization_map
        self._process = multiprocessing.Process(target=self._execute, daemon=True)

    def _send_message(self, message_type: str, value: dict[Any, Any] | str) -> None:
        self._messages_queue.put(Message(message_type, self._id, value))

    def _send_exception(self, exception: BaseException) -> None:
        backtrace = get_backtrace_info(exception)
        self._send_message(message_type_runtime_error, create_runtime_error_description(exception.__str__(), backtrace))

    def save_placeholder(self, placeholder_name: str, value: Any) -> None:
        """
        Save a calculated placeholder in the map.

        Parameters
        ----------
        placeholder_name : str
            Name of the placeholder.
        value : Any
            Actual value of the placeholder.
        """
        from safeds.data.image.containers import Image

        if isinstance(value, Image):
            import torch

            value = Image(value._image_tensor, torch.device("cpu"))
        self._placeholder_map[placeholder_name] = value
        placeholder_type = _get_placeholder_type(value)
        self._send_message(
            message_type_placeholder_type,
            create_placeholder_description(placeholder_name, placeholder_type),
        )

    def get_memoization_map(self) -> MemoizationMap:
        """
        Get the shared memoization map.

        Returns
        -------
        MemoizationMap
            Memoization Map
        """
        return self._memoization_map

    def _execute(self) -> None:
        logging.info(
            "Executing %s.%s.%s...",
            self._pipeline.main.modulepath,
            self._pipeline.main.module,
            self._pipeline.main.pipeline,
        )
        pipeline_finder = InMemoryFinder(self._pipeline.code)
        pipeline_finder.attach()
        main_module = f"gen_{self._pipeline.main.module}_{self._pipeline.main.pipeline}"
        # Populate current_pipeline global, so child process can save placeholders in correct location
        globals()["current_pipeline"] = self
        try:
            runpy.run_module(
                (
                    main_module
                    if len(self._pipeline.main.modulepath) == 0
                    else f"{self._pipeline.main.modulepath}.{main_module}"
                ),
                run_name="__main__",
                alter_sys=True,
            )
            self._send_message(message_type_runtime_progress, create_runtime_progress_done())
        except BaseException as error:  # noqa: BLE001
            self._send_exception(error)
        finally:
            pipeline_finder.detach()

    def execute(self) -> None:
        """
        Execute this pipeline in a newly created process.

        Results, progress and errors are communicated back to the main process.
        """
        self._process.start()


# Pipeline process object visible in child process
current_pipeline: PipelineProcess | None = None


def save_placeholder(placeholder_name: str, value: Any) -> None:
    """
    Save a placeholder for the current running pipeline.

    Parameters
    ----------
    placeholder_name : str
        Name of the placeholder.
    value : Any
        Actual value of the placeholder.
    """
    if current_pipeline is not None:
        current_pipeline.save_placeholder(placeholder_name, value)


def memoized_call(
    function_name: str,
    function_callable: typing.Callable,
    parameters: list[Any],
    hidden_parameters: list[Any],
) -> Any:
    """
    Call a function that can be memoized and save the result.

    If a function has been previously memoized, the previous result may be reused.

    Parameters
    ----------
    function_name : str
        Fully qualified function name
    function_callable : typing.Callable
        Function that is called and memoized if the result was not found in the memoization map
    parameters : list[Any]
        List of parameters for the function
    hidden_parameters : list[Any]
        List of hidden parameters for the function. This is used for memoizing some impure functions.

    Returns
    -------
    Any
        The result of the specified function, if any exists
    """
    if current_pipeline is None:
        return None  # pragma: no cover
    memoization_map = current_pipeline.get_memoization_map()
    return memoization_map.memoized_function_call(function_name, function_callable, parameters, hidden_parameters)


def file_mtime(filename: str) -> int | None:
    """
    Get the last modification timestamp of the provided file.

    Parameters
    ----------
    filename: str
        Name of the file

    Returns
    -------
    int | None
        Last modification timestamp if the provided file exists, otherwise None
    """
    try:
        return Path(filename).stat().st_mtime_ns
    except FileNotFoundError:
        return None


def get_backtrace_info(error: BaseException) -> list[dict[str, Any]]:
    """
    Create a simplified backtrace from an exception.

    Parameters
    ----------
    error : BaseException
        Caught exception.

    Returns
    -------
    list[dict[str, Any]]
        List containing file and line information for each stack frame.
    """
    backtrace_list = []
    for frame in stack_data.core.FrameInfo.stack_data(error.__traceback__):
        backtrace_list.append({"file": frame.filename, "line": int(frame.lineno)})
    return backtrace_list


def _get_placeholder_type(value: Any) -> str:
    """
    Convert a python object to a Safe-DS type.

    Parameters
    ----------
    value : Any
        A python object.

    Returns
    -------
    str
        Safe-DS name corresponding to the given python object instance.
    """
    match value:
        case bool():
            return "Boolean"
        case float():
            return "Float"
        case int():
            return "Int"
        case str():
            return "String"
        case object():
            object_name = type(value).__name__
            match object_name:
                case "function":
                    return "Callable"
                case "NoneType":
                    return "Null"
                case _:
                    return object_name
        case _:  # pragma: no cover
            return "Any"  # pragma: no cover



"""Module containing JSON encoding utilities for Safe-DS types."""

import base64
import json
import math
from typing import Any


class SafeDsEncoder(json.JSONEncoder):
    """JSON Encoder for custom Safe-DS types."""

    def default(self, o: Any) -> Any:
        """
        Convert specific Safe-DS types to a JSON-serializable representation.

        If values are custom Safe-DS types (such as Table or Image) they are converted to a serializable representation.
        If a value is not handled here, the default encoding implementation is called.
        In case of Tables, note that NaN values are converted to JSON null values.

        Parameters
        ----------
        o: Any
            An object that needs to be encoded to JSON.

        Returns
        -------
        Any
            The passed object represented in a way that is serializable to JSON.
        """
        # Moving these imports to the top drastically increases startup time
        from safeds.data.image.containers import Image
        from safeds.data.tabular.containers import Table

        if isinstance(o, Table):
            dict_with_nan_infinity = o.to_dict()
            # Convert NaN / Infinity to None, as the JSON encoder generates invalid JSON otherwise
            return {
                key: [
                    value if not isinstance(value, float) or math.isfinite(value) else None
                    for value in dict_with_nan_infinity[key]
                ]
                for key in dict_with_nan_infinity
            }
        if isinstance(o, Image):
            # Send images together with their format, by default images are encoded only as PNG
            return {"format": "png", "bytes": str(base64.encodebytes(o._repr_png_()), "utf-8")}
        return json.JSONEncoder.default(self, o)



"""Module that contains the infrastructure for finding and loading modules in-memory."""

import importlib.abc
import importlib.util
import logging
import sys
import types
import typing
from abc import ABC
from importlib.machinery import ModuleSpec


class InMemoryLoader(importlib.abc.SourceLoader, ABC):
    """Load a virtual python module from a byte array and a filename."""

    def __init__(self, code_bytes: bytes, filename: str):
        """
        Create a new in-memory loader.

        Parameters
        ----------
        code_bytes : bytes
            Byte array containing python code.
        filename : str
            Filename of the python module.
        """
        self.code_bytes = code_bytes
        self.filename = filename

    def get_data(self, _path: bytes | str) -> bytes:
        """
        Get module code as a byte array.

        Parameters
        ----------
        _path : bytes | str
            Module path.

        Returns
        -------
        bytes
            Module code.
        """
        return self.code_bytes

    def get_filename(self, _fullname: str) -> str:
        """
        Get file name for a module path.

        Parameters
        ----------
        _fullname : str
            Module path.

        Returns
        -------
        str
            virtual module path, as located in the code array in the InMemoryFinder that created this loader.
        """
        return self.filename


class InMemoryFinder(importlib.abc.MetaPathFinder):
    """Find python modules in an in-memory dictionary."""

    def __init__(self, code: dict[str, dict[str, str]]):
        """
        Create a new in-memory finder.

        Parameters
        ----------
        code : dict[str, dict[str, str]]
            A dictionary containing the code to be executed,
            grouped by module path containing a mapping from module name to module code.
        """
        self.code = code
        self.allowed_packages = set(code.keys())
        self.imports_to_remove: set[str] = set()
        for key in code:
            self._add_possible_packages_for_package_path(key)

    def _add_possible_packages_for_package_path(self, package_path: str) -> None:
        while "." in package_path:
            package_path = package_path.rpartition(".")[0]
            self.allowed_packages.add(package_path)

    def find_spec(
        self,
        fullname: str,
        path: typing.Sequence[str] | None = None,
        target: types.ModuleType | None = None,
    ) -> ModuleSpec | None:
        """
        Find a module which may be registered in the code dictionary.

        Parameters
        ----------
        fullname : str
            Full module path (separated with '.').
        path : typing.Sequence[str] | None
            Module Path.
        target : types.ModuleType | None
            Module Type.

        Returns
        -------
        ModuleSpec | None
            A module spec, if found. None otherwise
        """
        logging.debug("Find Spec: %s %s %s", fullname, path, target)
        if fullname in self.allowed_packages:
            parent_package = importlib.util.spec_from_loader(
                fullname,
                loader=InMemoryLoader(b"", fullname.replace(".", "/")),
            )
            if parent_package is None:
                return None  # pragma: no cover
            if parent_package.submodule_search_locations is None:
                parent_package.submodule_search_locations = []
            parent_package.submodule_search_locations.append(fullname.replace(".", "/"))
            self.imports_to_remove.add(fullname)
            return parent_package
        module_path = fullname.split(".")
        if len(module_path) == 1 and "" in self.code and fullname in self.code[""]:
            self.imports_to_remove.add(fullname)
            return importlib.util.spec_from_loader(
                fullname,
                loader=InMemoryLoader(self.code[""][fullname].encode("utf-8"), fullname.replace(".", "/")),
                origin="",
            )
        parent_package_path = ".".join(module_path[:-1])
        submodule_name = module_path[-1]
        if parent_package_path in self.code and submodule_name in self.code[parent_package_path]:
            self.imports_to_remove.add(fullname)
            return importlib.util.spec_from_loader(
                fullname,
                loader=InMemoryLoader(
                    self.code[parent_package_path][submodule_name].encode("utf-8"),
                    fullname.replace(".", "/"),
                ),
                origin=parent_package_path,
            )
        return None  # pragma: no cover

    def attach(self) -> None:
        """Attach this finder to the meta path."""
        sys.meta_path.append(self)

    def detach(self) -> None:
        """Remove modules found in this finder and remove finder from meta path."""
        # As modules should not be used from other modules, outside our pipeline,
        # it should be safe to just remove all newly imported modules
        for key in self.imports_to_remove:
            if key in sys.modules:
                del sys.modules[key]
        sys.meta_path.remove(self)



import argparse
import logging
from importlib.metadata import version

from safeds_runner.server.main import start_server


class Commands:
    START = "start"


def cli() -> None:  # pragma: no cover
    """Run the application via the command line."""
    args = _get_args()

    # Set logging level
    if args.verbose:
        logging.basicConfig(level=logging.DEBUG)
    else:
        logging.basicConfig(level=logging.INFO)

    # Run command
    match args.command:
        case Commands.START:
            start_server(args.port)


def _get_args() -> argparse.Namespace:  # pragma: no cover
    parser = argparse.ArgumentParser(description="Execute Safe-DS programs that were compiled to Python.")
    parser.add_argument(
        "-V",
        "--version",
        action="version",
        version=f"%(prog)s {version('safe-ds-runner')}",
    )
    parser.add_argument("-v", "--verbose", help="increase logging verbosity", action="store_true")

    # Commands
    subparsers = parser.add_subparsers(dest="command")
    _add_start_subparser(subparsers)

    return parser.parse_args()


def _add_start_subparser(subparsers: argparse._SubParsersAction) -> None:  # pragma: no cover
    parser = subparsers.add_parser(Commands.START, help="start the Safe-DS Runner server")
    parser.add_argument("-p", "--port", type=int, default=5000, help="the port to use")



"""Module containing the main entry point, for starting the Safe-DS runner."""

import logging

from ._server import SafeDsServer


def start_server(port: int) -> None:
    """Start the Safe-DS Runner server."""
    # Allow prints to be unbuffered by default
    import builtins
    import functools

    builtins.print = functools.partial(print, flush=True)  # type: ignore[assignment]

    logging.getLogger().setLevel(logging.DEBUG)

    safeds_server = SafeDsServer()
    safeds_server.listen(port)  # pragma: no cover
